using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using PVOID = System.IntPtr;
using DWORD = System.UInt32;

namespace usb_api
{
    /// <summary>
    /// Generic interface to a USB PIC using normal VID/PID
    /// </summary>
	unsafe public class usb_interface
	{
		#region  String Definitions of Pipes and VID_PID
        /// <summary>
        /// VID/PID for PIC running Microchip default USB bootloader firmware
        /// </summary>
		public string vid_pid_boot= "vid_04d8&pid_000b";
        /// <summary>
        /// VID/PID for PIC running generic firmware
        /// </summary>
		public string vid_pid_norm= "vid_04d8&pid_000c";

        /// <summary>
        /// OUT endpoint on PIC
        /// </summary>
		protected string out_pipe= "\\MCHP_EP1";
        /// <summary>
        /// IN endpoint on PIC (synchronous)
        /// </summary>
		protected string in_pipe= "\\MCHP_EP1";
        /// <summary>
        /// IN endpoint on PIC (asynchronous)
        /// </summary>
        protected string in_pipe_async = "\\MCHP_EP1_ASYNC";
		#endregion

		#region Imported DLL functions from mpusbapi.dll
		[DllImport("mpusbapi.dll")]
		private static extern DWORD _MPUSBGetDLLVersion();
		[DllImport("mpusbapi.dll")]
		private static extern DWORD _MPUSBGetDeviceCount(string pVID_PID);
        /// <summary>
        /// Returns the handle to the endpoint pipe with matching VID/PID
        /// </summary>
        /// <param name="instance">An instance number of the device to open.  Typically zero for a single device</param>
        /// <param name="pVID_PID">A string containing the PID/VID value of the target device</param>
        /// <param name="pEP">A string of the endpoint number on the target endpoint to open</param>
        /// <param name="dwDir">Specifies the direction of the endpoint (0=OUT, 1=IN)</param>
        /// <param name="dwReserved">n/a</param>
        /// <returns>Handle to the endpoint pipe</returns>
		[DllImport("mpusbapi.dll")]
		protected static extern void* _MPUSBOpen(DWORD instance,string pVID_PID,string pEP,DWORD dwDir,DWORD dwReserved);
		/// <summary>
		/// Attempts to read from specified pipe.
		/// </summary>
		/// <param name="handle">Handle to the pipe</param>
		/// <param name="pData">Pointer to buffer for received data</param>
		/// <param name="dwLen">Number of bytes to read from pipe</param>
		/// <param name="pLength">Pointer to value of number of bytes read</param>
		/// <param name="dwMilliseconds">Timeout interval in milliseconds</param>
		/// <returns>0=failure, 1=success</returns>
        [DllImport("mpusbapi.dll")]
		protected static extern DWORD _MPUSBRead(void* handle,void* pData,DWORD dwLen,DWORD* pLength,DWORD dwMilliseconds);
		[DllImport("mpusbapi.dll")]
		private static extern DWORD _MPUSBWrite(void* handle,void* pData,DWORD dwLen,DWORD* pLength,DWORD dwMilliseconds);
		[DllImport("mpusbapi.dll")]
        private static extern DWORD _MPUSBReadInt(void* handle, void* pData, DWORD dwLen, DWORD* pLength, DWORD dwMilliseconds);
		[DllImport("mpusbapi.dll")]
		private static extern bool _MPUSBClose(void* handle);
		#endregion
        
        /// <summary>
        /// OUT endpoint pipe
        /// </summary>
		protected void* myOutPipe;
        /// <summary>
        /// IN endpoint pipe (synchronous)
        /// </summary>
		protected void* myInPipe;
        /// <summary>
        /// IN endpoint pipe (asynchronous)
        /// </summary>
        protected void* myAsyncInPipe;

        /// <summary>
        /// Represents whether pipes have been opened
        /// </summary>
        protected bool prvPipesOpen;
        /// <summary>
        /// Represents whether pipes have been opened
        /// </summary>
        public bool PipesOpen
        {
            get { return prvPipesOpen; }
            set { }
        }

        /// <summary>
        /// Commands recognized by generic firmware.  Custom commands may be added or declared elsewhere.
        /// </summary>
        public enum Commands : byte
        {
            /// <summary>
            /// Turn LED 3 or 4 on demo board on or off.
            /// </summary>
            UPDATE_LED = 0x32,
            /// <summary>
            /// Instructs PIC to reset
            /// </summary>
            RESET = 0xFF
        }

        /// <summary>
        /// Opens USB pipes to the PIC using the normal VID/PID
        /// </summary>
		protected virtual void OpenPipes()
		{
			DWORD selection=0; // Selects the device to connect to
			myOutPipe = _MPUSBOpen(selection,vid_pid_norm,out_pipe,0,0);
			myInPipe = _MPUSBOpen(selection,vid_pid_norm,in_pipe,1,0);
            prvPipesOpen = true;
		}

        /// <summary>
        /// Constructor
        /// </summary>
        public usb_interface()
        {
            prvPipesOpen = false;
        }

        /// <summary>
        /// Closes all open pipes.  Should be called if a different instance of the class needs access
        /// </summary>
		public void ClosePipes()
		{
			_MPUSBClose(myOutPipe);
            _MPUSBClose(myAsyncInPipe);
			_MPUSBClose(myInPipe);
            prvPipesOpen = false;
		}

        /// <summary>
        /// Writes one packet to PIC, receives response packet.
        /// </summary>
        /// <param name="SendData">Pointer to bytes to be sent</param>
        /// <param name="SendLength">Number of bytes to be sent</param>
        /// <param name="ReceiveData">Pointer to location for received data</param>
        /// <param name="ReceiveLength">Expected number of received bytes</param>
        /// <returns>1=Success, 2=Incorrect RX Length, 3=Read Failed</returns>
		protected DWORD SendReceivePacket(byte* SendData, DWORD SendLength, byte* ReceiveData, DWORD *ReceiveLength)
		{
            if (prvPipesOpen==false)
                OpenPipes();
            uint SendDelay = 1000;
            uint ReceiveDelay = 1000;
            DWORD rval = 0;
            DWORD SentDataLength;
            DWORD ExpectedReceiveLength = *ReceiveLength;
            if (_MPUSBWrite(myOutPipe, (void*)SendData, SendLength, &SentDataLength, SendDelay) == 1)
            {

                if (_MPUSBRead(myInPipe, (void*)ReceiveData, ExpectedReceiveLength, ReceiveLength, ReceiveDelay) == 1)
                {
                    if (*ReceiveLength == ExpectedReceiveLength)
                    {
                        rval = 1;   // Success!
                    }
                    else if (*ReceiveLength < ExpectedReceiveLength)
                    {
                        rval = 2;   // Partially failed, incorrect receive length
                    }
                }
                else
                {
                    rval = 3;   //Read Failed
                }
            }
            return rval;
		}

        /// <summary>
        /// Gets the version of mpusbapi.dll
        /// </summary>
        /// <returns>Version number</returns>
		public DWORD GetDLLVersion()
		{
			return _MPUSBGetDLLVersion();
		}
        /// <summary>
        /// Scans the USB bus for connected devices with matching VID/PID
        /// </summary>
        /// <param name="Vid_Pid">VID/PID to scan for</param>
        /// <returns>Number of connected devices</returns>
		public DWORD GetDeviceCount(string Vid_Pid)
		{
			return _MPUSBGetDeviceCount(Vid_Pid);
		}

        /// <summary>
        /// Simple method for sending/receiving USB data from the PIC.
        /// </summary>
        /// <param name="Command">Command byte</param>
        /// <param name="rxlength">Expected length of received packet</param>
        /// <param name="data">Array of bytes to send after command - pass null if not applicable</param>
        /// <param name="dataout">Array for received bytes</param>
        /// <returns>1=Success, 2=Incorrect RX Length, 3=Read Failed</returns>
        public uint EasyCommand(byte Command, int rxlength, byte[] data, out byte[] dataout)
        {            
            byte* send_buf = stackalloc byte[64];
            byte* receive_buf = stackalloc byte[64];
            DWORD RecvLength;
            uint rval;

            RecvLength = (DWORD)rxlength;
            send_buf[0] = Command;
            if (data != null)
            {
                for (int i = 0; i < data.Length; i++)
                { send_buf[i + 1] = data[i]; }
            }
            else
                data = new byte[] { }; //just set to empty array so .Length member is valid at 0.
            rval = SendReceivePacket(send_buf, 1 + (uint)data.Length, receive_buf, &RecvLength);
            if (rval != 1)
            { dataout = null; return rval; }
            dataout = new byte[rxlength];
            for (int i = 0; i < rxlength; i++)
            { dataout[i] = receive_buf[i]; }
            return 1;
        }
	}
}
